/* Doubly linked-list implementation for containing items of type ListItem.
 *
 * Notes:
 *    This implementation demonstrates the use of Arenas for centralised memory 
 * management and Free-lists for caching of released nodes.
 *    (1) The Arena centralises and controls the allocation of memory for all lists.
 * In particular, it prevents memory leaks by providing a single final deletion of 
 * all list memory. It allocates memory in chunks, which improves locality-of-reference
 * and reduces the space overhead incured by global heap allocation (ie, the internal 
 * record keeping for each object allocated by malloc() etc).
 *    (2) The Free-list is a list of released List nodes that are available for
 * allocation to other lists. This node cache is faster than global memory allocation
 * because (i) it is not thread-safe, and (ii) it manages objects of fixed-size 
 * (whereas the global heap allocator is general purpose). The free-list is particularly 
 * well suited to lists of small objects that are created and deleted frequently.
 *    (3) A non-usable address ADDR_DEBUG is used to mark all free-list nodes. This 
 * catches certain client bugs where a list node is deleted, but the client code
 * continues to use the node pointer.
 *
 * Tim Bailey 2004.
 */

#include <stdlib.h>
#include <assert.h>
#include "dlist.h"

#define CHUNK       10              /* Number of List nodes allocated with each access of global heap. */
#define ADDR_DEBUG  ((void *)0x01)  /* Non-usable address to mark free-list nodes. */
typedef struct ListArray ListArray; /* List of array of Lists, managed by the Arena. */

struct List { /* underlying list-node type */
    ListItem item;
    List *next;
    List *prev;
};

struct ListArray { /* list-array node type: used by arena */
    List chunk[CHUNK];
    ListArray *next;
};

static ListArray *arena = NULL; /* arena manages ALL list memory */
static List *freelist = NULL;   /* freelist manages non-used List nodes */

static List* allocate_nodearray(void)
/* Add a new ListArray object to the arena linked list (which manages all list
 * node memory). Return a pointer to the new link-list-array (with CHUNK elements) if
 * successful, or NULL if fails. */
{
    ListArray *node = (ListArray *) malloc(sizeof(ListArray));
    if (!node)
        return NULL;

    { /* Link array elements into a singly-linked list */
        List *i = node->chunk;
        List *end = &node->chunk[CHUNK-1]; 
        for (; i!=end; ++i) {
            i->next = i + 1;
            i->prev = ADDR_DEBUG; 
        }
        i->next = NULL;
        i->prev = ADDR_DEBUG; 
    }

    node->next = arena; /* add new array to head of arena */
    arena = node;
    return node->chunk;
}

static List* allocate_node(void)
/* Allocate memory from freelist. If freelist empty, extend it via the arena. 
 * Returns pointer to node if successful, NULL if fails. */
{
    List *head;

    if (freelist == NULL) { /* freelist empty */        
        freelist = allocate_nodearray(); 
        if (!freelist)
            return NULL;
    }

    /* Shift freelist head along by one, and return old head */
    head = freelist;
    freelist = freelist->next;
    return head;
}

static void deallocate_node(List *node)
/* Return node to head of freelist */
{
    node->next = freelist;
    node->prev = ADDR_DEBUG; 
    freelist = node;
}

static List *create_node(ListItem item)
/* Create and initialise new node. Return node if successful, or NULL if fails. */
{
    List *node = allocate_node();

    if (node) {
        node->item = item;
        node->next = NULL;
        node->prev = NULL;
    }    
    return node;
}

List *list_insert_front(List *node, ListItem item)
/* Add new item to the head of the list.
 * Return new head if successful, or NULL if fails. */
{
    List *newnode = create_node(item);

    if (newnode && node) { 
        assert(node->prev != ADDR_DEBUG);
        node = list_get_head(node);
        node->prev = newnode;
        newnode->next = node;
    }
    return newnode;
}

List *list_insert_back(List *node, ListItem item)
/* Add new item to the tail of the list.
 * Return new tail if successful, or NULL if fails. */
{
    List *newnode = create_node(item);

    if (newnode && node) {
        assert(node->prev != ADDR_DEBUG);
        node = list_get_tail(node);
        node->next = newnode;
        newnode->prev = node;
    }
    return newnode;
}

List *list_insert_after(List *node, ListItem item)
/* Add new item to the next position in the list after node. 
 * Return new node if successful, or NULL if fails. */
{
    List *newnode = create_node(item);

    if (newnode && node) { /* splice new node into list. */
        assert(node->prev != ADDR_DEBUG);
        if (node->next)
            node->next->prev = newnode;
        newnode->next = node->next; 
        node->next = newnode;
        newnode->prev = node;
    }
    return newnode;
}

List *list_insert_before(List *node, ListItem item)
/* Add new item to the position in the list before node. 
 * Return new node if successful, or NULL if fails. */
{
    List *newnode = create_node(item);

    if (newnode && node) { /* splice new node into list. */
        assert(node->prev != ADDR_DEBUG);
        if (node->prev)
            node->prev->next = newnode;
        newnode->prev = node->prev;
        node->prev = newnode;
        newnode->next = node; 
    }
    return newnode;
}

void list_delete_node(List *node)
/* Remove specified node from list. */
{
    assert(node->prev != ADDR_DEBUG);

    /* reconnect list on either side of node. */
    if (node->prev) node->prev->next = node->next; 
    if (node->next) node->next->prev = node->prev;
    deallocate_node(node); /* return node to freelist */
}

List *list_delete_after(List *node)
/* Remove node and all nodes following from list. Return node before node (ie, new tail). */
{
    List *before = node->prev;
    assert(node->prev != ADDR_DEBUG);

    if (before) 
        before->next = NULL;
    while (node) {
        List *tmp = node;
        node = node->next;
        deallocate_node(tmp);
    }
    return before;
}

List *list_delete_before(List *node)
/* Remove node and all nodes preceeding it from list. Return node after node (ie, new head). */
{
    List *after = node->next;
    assert(node->prev != ADDR_DEBUG);

    if (after)
        after->prev = NULL;
    while (node) {
        List *tmp = node;
        node = node->prev;
        deallocate_node(tmp);
    }
    return after;
}

void list_delete_all_lists(void)
/* Release all list memory as managed by the Arena */
{
    while (arena) {
        ListArray *tmp = arena;
        arena = arena->next;
        free(tmp);
    }
}

List *list_get_head(List *node)
/* Returns head of list */
{
    assert(node->prev != ADDR_DEBUG);

    while (node->prev)
        node = node->prev;
    return node;
}

List *list_get_tail(List *node)
/* Returns tail of list */
{
    assert(node->prev != ADDR_DEBUG);

    while (node->next)
        node = node->next;
    return node;
}

List *list_get_next(List *node) { assert(node->prev != ADDR_DEBUG); return node->next; }
List *list_get_prev(List *node) { assert(node->prev != ADDR_DEBUG); return node->prev; }
ListItem *list_get_item(List *node) { assert(node->prev != ADDR_DEBUG); return &node->item; }
